@@ -17,26 +17,23 @@
 # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
+"""Generic utility helpers."""
 
-from collections import OrderedDict
-import glob
-import imp
-from itertools import product
+import contextlib
+import inspect
+import logging
 import os
-try:
-    from pathlib import Path
-except ImportError:
-    from pathlib2 import Path
+import pprint
 import subprocess
+from argparse import Namespace
+from collections import OrderedDict
+from functools import lru_cache
+from pathlib import Path
+from typing import Callable, ItemsView, List, Optional, Tuple
 
-import six
 import yaml
-from yaml.composer import Composer
-import ruamel.yaml
-
 from ansible import constants
-from ansible.errors import AnsibleError
-from ansible.errors import AnsibleParserError
+from ansible.errors import AnsibleError, AnsibleParserError
 from ansible.parsing.dataloader import DataLoader
 from ansible.parsing.mod_args import ModuleArgsParser
 from ansible.parsing.splitter import split_args
@@ -45,22 +42,32 @@ from ansible.parsing.yaml.loader import AnsibleLoader
 from ansible.parsing.yaml.objects import AnsibleSequence
 from ansible.plugins.loader import module_loader
 from ansible.template import Templar
+from yaml.composer import Composer
+from yaml.representer import RepresenterError
 
+from ansiblelint.constants import ANSIBLE_FAILURE_RC, FileType
+from ansiblelint.errors import MatchError
+from ansiblelint.file_utils import normpath
 
 # ansible-lint doesn't need/want to know about encrypted secrets, so we pass a
 # string as the password to enable such yaml files to be opened and parsed
 # successfully.
 DEFAULT_VAULT_PASSWORD = 'x'
 
+PLAYBOOK_DIR = os.environ.get('ANSIBLE_PLAYBOOK_DIR', None)
 
-def parse_yaml_from_file(filepath):
+
+_logger = logging.getLogger(__name__)
+
+
+def parse_yaml_from_file(filepath: str) -> dict:
     dl = DataLoader()
     if hasattr(dl, 'set_vault_password'):
         dl.set_vault_password(DEFAULT_VAULT_PASSWORD)
     return dl.load_from_file(filepath)
 
 
-def path_dwim(basedir, given):
+def path_dwim(basedir: str, given: str) -> str:
     dl = DataLoader()
     dl.set_basedir(basedir)
     return dl.path_dwim(given)
@@ -98,24 +105,6 @@ BLOCK_NAME_TO_ACTION_TYPE_MAP = {
 }
 
 
-def load_plugins(directory):
-    result = []
-    fh = None
-
-    for pluginfile in glob.glob(os.path.join(directory, '[A-Za-z]*.py')):
-
-        pluginname = os.path.basename(pluginfile.replace('.py', ''))
-        try:
-            fh, filename, desc = imp.find_module(pluginname, [directory])
-            mod = imp.load_module(pluginname, fh, filename, desc)
-            obj = getattr(mod, pluginname)()
-            result.append(obj)
-        finally:
-            if fh:
-                fh.close()
-    return result
-
-
 def tokenize(line):
     tokens = line.lstrip().split(" ")
     if tokens[0] == '-':
@@ -137,7 +126,7 @@ def tokenize(line):
     return (command, args, kwargs)
 
 
-def _playbook_items(pb_data):
+def _playbook_items(pb_data: dict) -> ItemsView:
     if isinstance(pb_data, dict):
         return pb_data.items()
     elif not pb_data:
@@ -146,7 +135,17 @@ def _playbook_items(pb_data):
         return [item for play in pb_data for item in play.items()]
 
 
-def find_children(playbook, playbook_dir):
+def _rebind_match_filename(filename: str, func) -> Callable:
+    def func_wrapper(*args, **kwargs):
+        try:
+            return func(*args, **kwargs)
+        except MatchError as e:
+            e.filename = filename
+            raise e
+    return func_wrapper
+
+
+def find_children(playbook: Tuple[str, str], playbook_dir: str) -> List:
     if not os.path.exists(playbook[0]):
         return []
     if playbook[1] == 'role':
@@ -160,7 +159,8 @@ def find_children(playbook, playbook_dir):
     basedir = os.path.dirname(playbook[0])
     items = _playbook_items(playbook_ds)
     for item in items:
-        for child in play_children(basedir, item, playbook[1], playbook_dir):
+        for child in _rebind_match_filename(playbook[0], play_children)(
+                basedir, item, playbook[1], playbook_dir):
             if "$" in child['path'] or "{{" in child['path']:
                 continue
             valid_tokens = list()
@@ -182,7 +182,7 @@ def template(basedir, value, vars, fail_on_undefined=False, **kwargs):
                                  **dict(kwargs, fail_on_undefined=fail_on_undefined))
         # Hack to skip the following exception when using to_json filter on a variable.
         # I guess the filter doesn't like empty vars...
-    except (AnsibleError, ValueError):
+    except (AnsibleError, ValueError, RepresenterError):
         # templating failed, so just keep value as is.
         pass
     return value
@@ -210,7 +210,7 @@ def play_children(basedir, item, parent_type, playbook_dir):
         if v:
             v = template(os.path.abspath(basedir),
                          v,
-                         dict(playbook_dir=os.path.abspath(basedir)),
+                         dict(playbook_dir=PLAYBOOK_DIR or os.path.abspath(basedir)),
                          fail_on_undefined=False)
             return delegate_map[k](basedir, k, v, parent_type)
     return []
@@ -221,55 +221,73 @@ def _include_children(basedir, k, v, parent_type):
     (command, args, kwargs) = tokenize("{0}: {1}".format(k, v))
 
     result = path_dwim(basedir, args[0])
-    if not os.path.exists(result) and not basedir.endswith('tasks'):
-        result = path_dwim(os.path.join(basedir, '..', 'tasks'), v)
+    if not os.path.exists(result):
+        result = path_dwim(os.path.join(os.path.dirname(basedir)), v)
     return [{'path': result, 'type': parent_type}]
 
 
-def _taskshandlers_children(basedir, k, v, parent_type):
+def _taskshandlers_children(basedir, k, v, parent_type: FileType) -> List:
     results = []
     for th in v:
-        if 'include' in th:
-            append_children(th['include'], basedir, k, parent_type, results)
-        elif 'include_tasks' in th:
-            append_children(th['include_tasks'], basedir, k, parent_type, results)
-        elif 'import_playbook' in th:
-            append_children(th['import_playbook'], basedir, k, parent_type, results)
-        elif 'import_tasks' in th:
-            append_children(th['import_tasks'], basedir, k, parent_type, results)
-        elif 'import_role' in th:
+        with contextlib.suppress(LookupError):
+            children = _get_task_handler_children_for_tasks_or_playbooks(
+                th, basedir, k, parent_type,
+            )
+            results.append(children)
+            continue
+
+        if 'include_role' in th or 'import_role' in th:  # lgtm [py/unreachable-statement]
             th = normalize_task_v2(th)
-            results.extend(_roles_children(basedir, k, [th['action'].get('name')], parent_type,
-                                           main=th['action'].get('tasks_from', 'main')))
-        elif 'include_role' in th:
-            th = normalize_task_v2(th)
-            results.extend(_roles_children(basedir, k, [th['action'].get('name')],
+            _validate_task_handler_action_for_role(th['action'])
+            results.extend(_roles_children(basedir, k, [th['action'].get("name")],
                                            parent_type,
                                            main=th['action'].get('tasks_from', 'main')))
-        elif 'block' in th:
-            results.extend(_taskshandlers_children(basedir, k, th['block'], parent_type))
-            if 'rescue' in th:
-                results.extend(_taskshandlers_children(basedir, k, th['rescue'], parent_type))
-            if 'always' in th:
-                results.extend(_taskshandlers_children(basedir, k, th['always'], parent_type))
+            continue
+
+        if 'block' not in th:
+            continue
+
+        results.extend(_taskshandlers_children(basedir, k, th['block'], parent_type))
+        if 'rescue' in th:
+            results.extend(_taskshandlers_children(basedir, k, th['rescue'], parent_type))
+        if 'always' in th:
+            results.extend(_taskshandlers_children(basedir, k, th['always'], parent_type))
+
     return results
 
 
-def append_children(taskhandler, basedir, k, parent_type, results):
-    # when taskshandlers_children is called for playbooks, the
-    # actual type of the included tasks is the section containing the
-    # include, e.g. tasks, pre_tasks, or handlers.
-    if parent_type == 'playbook':
-        playbook_section = k
-    else:
-        playbook_section = parent_type
-    results.append({
-        'path': path_dwim(basedir, taskhandler),
-        'type': playbook_section
-    })
+def _get_task_handler_children_for_tasks_or_playbooks(
+        task_handler, basedir: str, k, parent_type: FileType,
+) -> dict:
+    """Try to get children of taskhandler for include/import tasks/playbooks."""
+    child_type = k if parent_type == 'playbook' else parent_type
+    task_include_keys = 'include', 'include_tasks', 'import_playbook', 'import_tasks'
+    for task_handler_key in task_include_keys:
+        with contextlib.suppress(KeyError):
+            return {
+                'path': path_dwim(basedir, task_handler[task_handler_key]),
+                'type': child_type,
+            }
+
+    raise LookupError(
+        f'The node contains none of: {", ".join(task_include_keys)}',
+    )
 
 
-def _roles_children(basedir, k, v, parent_type, main='main'):
+def _validate_task_handler_action_for_role(th_action: dict) -> None:
+    """Verify that the task handler action is valid for role include."""
+    module = th_action['__ansible_module__']
+
+    if 'name' not in th_action:
+        raise MatchError(f"Failed to find required 'name' key in {module!s}")
+
+    if not isinstance(th_action['name'], str):
+        raise RuntimeError(
+            f"Value assigned to 'name' key on '{module!s}' is not a string.",
+        )
+
+
+def _roles_children(basedir: str, k, v, parent_type: FileType, main='main') -> list:
     results = []
     for role in v:
         if isinstance(role, dict):
@@ -286,12 +304,12 @@ def _roles_children(basedir, k, v, parent_type, main='main'):
     return results
 
 
-def _load_library_if_exists(path):
+def _load_library_if_exists(path: str) -> None:
     if os.path.exists(path):
         module_loader.add_directory(path)
 
 
-def _rolepath(basedir, role):
+def _rolepath(basedir: str, role: str) -> Optional[str]:
     role_path = None
 
     possible_paths = [
@@ -307,7 +325,7 @@ def _rolepath(basedir, role):
 
     if constants.DEFAULT_ROLES_PATH:
         search_locations = constants.DEFAULT_ROLES_PATH
-        if isinstance(search_locations, six.string_types):
+        if isinstance(search_locations, str):
             search_locations = search_locations.split(os.pathsep)
         for loc in search_locations:
             loc = os.path.expanduser(loc)
@@ -326,7 +344,7 @@ def _rolepath(basedir, role):
     return role_path
 
 
-def _look_for_role_files(basedir, role, main='main'):
+def _look_for_role_files(basedir: str, role: str, main='main') -> list:
     role_path = _rolepath(basedir, role)
     if not role_path:
         return []
@@ -356,34 +374,59 @@ def rolename(filepath):
 
 def _kv_to_dict(v):
     (command, args, kwargs) = tokenize(v)
-    return (dict(__ansible_module__=command, __ansible_arguments__=args, **kwargs))
+    return dict(__ansible_module__=command, __ansible_arguments__=args, **kwargs)
 
 
-def normalize_task_v2(task):
+def _sanitize_task(task: dict) -> dict:
+    """Return a stripped-off task structure compatible with new Ansible.
+
+    This helper takes a copy of the incoming task and drops
+    any internally used keys from it.
+    """
+    result = task.copy()
+    # task is an AnsibleMapping which inherits from OrderedDict, so we need
+    # to use `del` to remove unwanted keys.
+    for k in ['skipped_rules', FILENAME_KEY, LINE_NUMBER_KEY]:
+        if k in result:
+            del result[k]
+    return result
+
+
+# FIXME: drop noqa once this function is made simpler
+# Ref: https://github.com/ansible/ansible-lint/issues/744
+def normalize_task_v2(task: dict) -> dict:  # noqa: C901
     """Ensure tasks have an action key and strings are converted to python objects."""
     result = dict()
-    mod_arg_parser = ModuleArgsParser(task)
+    if 'always_run' in task:
+        # FIXME(ssbarnea): Delayed import to avoid circular import
+        # See https://github.com/ansible/ansible-lint/issues/880
+        # noqa: # pylint:disable=cyclic-import,import-outside-toplevel
+        from ansiblelint.rules.AlwaysRunRule import AlwaysRunRule
+
+        raise MatchError(
+            rule=AlwaysRunRule,
+            filename=task[FILENAME_KEY],
+            linenumber=task[LINE_NUMBER_KEY])
+
+    sanitized_task = _sanitize_task(task)
+    mod_arg_parser = ModuleArgsParser(sanitized_task)
     try:
         action, arguments, result['delegate_to'] = mod_arg_parser.parse()
     except AnsibleParserError as e:
         try:
             task_info = "%s:%s" % (task[FILENAME_KEY], task[LINE_NUMBER_KEY])
-            del task[FILENAME_KEY]
-            del task[LINE_NUMBER_KEY]
         except KeyError:
             task_info = "Unknown"
-        try:
-            import pprint
-            pp = pprint.PrettyPrinter(indent=2)
-            task_pprint = pp.pformat(task)
-        except ImportError:
-            task_pprint = task
-        raise SystemExit("Couldn't parse task at %s (%s)\n%s" % (task_info, e.message, task_pprint))
+        pp = pprint.PrettyPrinter(indent=2)
+        task_pprint = pp.pformat(sanitized_task)
+
+        _logger.critical("Couldn't parse task at %s (%s)\n%s", task_info, e.message, task_pprint)
+        raise SystemExit(ANSIBLE_FAILURE_RC)
 
     # denormalize shell -> command conversion
     if '_uses_shell' in arguments:
         action = 'shell'
-        del(arguments['_uses_shell'])
+        del arguments['_uses_shell']
 
     for (k, v) in list(task.items()):
         if k in ('action', 'local_action', 'args', 'delegate_to') or k == action:
@@ -397,19 +440,21 @@ def normalize_task_v2(task):
 
     if '_raw_params' in arguments:
         result['action']['__ansible_arguments__'] = arguments['_raw_params'].split(' ')
-        del(arguments['_raw_params'])
+        del arguments['_raw_params']
     else:
         result['action']['__ansible_arguments__'] = list()
 
     if 'argv' in arguments and not result['action']['__ansible_arguments__']:
         result['action']['__ansible_arguments__'] = arguments['argv']
-        del(arguments['argv'])
+        del arguments['argv']
 
     result['action'].update(arguments)
     return result
 
 
-def normalize_task_v1(task):
+# FIXME: drop noqa once this function is made simpler
+# Ref: https://github.com/ansible/ansible-lint/issues/744
+def normalize_task_v1(task):  # noqa: C901
     result = dict()
     for (k, v) in task.items():
         if k in VALID_KEYS or k.startswith('with_'):
@@ -421,7 +466,7 @@ def normalize_task_v1(task):
             else:
                 result[k] = v
         else:
-            if isinstance(v, six.string_types):
+            if isinstance(v, str):
                 v = _kv_to_dict(k + ' ' + v)
             elif not v:
                 v = dict(__ansible_module__=k)
@@ -450,17 +495,17 @@ def normalize_task_v1(task):
         #   module: ec2
         #   etc...
         result['action']['__ansible_module__'] = result['action']['module']
-        del(result['action']['module'])
+        del result['action']['module']
     if 'args' in result:
         result['action'].update(result.get('args'))
-        del(result['args'])
+        del result['args']
     return result
 
 
 def normalize_task(task, filename):
     ansible_action_type = task.get('__ansible_action_type__', 'task')
     if '__ansible_action_type__' in task:
-        del(task['__ansible_action_type__'])
+        del task['__ansible_action_type__']
     task = normalize_task_v2(task)
     task[FILENAME_KEY] = filename
     task['__ansible_action_type__'] = ansible_action_type
@@ -533,6 +578,7 @@ def get_normalized_tasks(yaml, file):
     return res
 
 
+@lru_cache(maxsize=128)
 def parse_yaml_linenumbers(data, filename):
     """Parse yaml as ansible.utils.parse_yaml but with linenumbers.
 
@@ -555,9 +601,8 @@ def parse_yaml_linenumbers(data, filename):
         return mapping
 
     try:
-        import inspect
         kwargs = {}
-        if 'vault_password' in inspect.getargspec(AnsibleLoader.__init__).args:
+        if 'vault_password' in inspect.getfullargspec(AnsibleLoader.__init__).args:
             kwargs['vault_password'] = DEFAULT_VAULT_PASSWORD
         loader = AnsibleLoader(data, **kwargs)
         loader.compose_node = compose_node
@@ -579,153 +624,7 @@ def get_first_cmd_arg(task):
     return first_cmd_arg
 
 
-def append_skipped_rules(pyyaml_data, file_text, file_type):
-    """Append 'skipped_rules' to individual tasks or single metadata block.
-
-    For a file, uses 2nd parser (ruamel.yaml) to pull comments out of
-    yaml subsets, check for '# noqa' skipped rules, and append any skips to the
-    original parser (pyyaml) data relied on by remainder of ansible-lint.
-
-    :param pyyaml_data: file text parsed via ansible and pyyaml.
-    :param file_text: raw file text.
-    :param file_type: type of file: tasks, handlers or meta.
-    :returns: original pyyaml_data altered with a 'skipped_rules' list added
-    to individual tasks, or added to the single metadata block.
-    """
-    try:
-        yaml_skip = _append_skipped_rules(pyyaml_data, file_text, file_type)
-    except RuntimeError as exc:
-        # Notify user of skip error, do not stop, do not change exit code
-        print('Error trying to append skipped rules: {!r}'.format(exc))
-        return pyyaml_data
-    return yaml_skip
-
-
-def _append_skipped_rules(pyyaml_data, file_text, file_type):
-    # parse file text using 2nd parser library
-    yaml = ruamel.yaml.YAML()
-    ruamel_data = yaml.load(file_text)
-
-    if file_type == 'meta':
-        pyyaml_data[0]['skipped_rules'] = _get_rule_skips_from_yaml(ruamel_data)
-        return pyyaml_data
-
-    # create list of blocks of tasks or nested tasks
-    if file_type in ('tasks', 'handlers'):
-        ruamel_task_blocks = ruamel_data
-        pyyaml_task_blocks = pyyaml_data
-    elif file_type in ('playbook', 'pre_tasks', 'post_tasks'):
-        try:
-            pyyaml_task_blocks = _get_task_blocks_from_playbook(pyyaml_data)
-            ruamel_task_blocks = _get_task_blocks_from_playbook(ruamel_data)
-        except (AttributeError, TypeError):
-            # TODO(awcrosby): running ansible-lint on any .yml file will
-            # assume it is a playbook, check needs to be added higher in the
-            # call stack, and can remove this except
-            return pyyaml_data
-    else:
-        raise RuntimeError('Unexpected file type: {}'.format(file_type))
-
-    # get tasks from blocks of tasks
-    pyyaml_tasks = _get_tasks_from_blocks(pyyaml_task_blocks)
-    ruamel_tasks = _get_tasks_from_blocks(ruamel_task_blocks)
-
-    # append skipped_rules for each task
-    for ruamel_task, pyyaml_task in zip(ruamel_tasks, pyyaml_tasks):
-        if pyyaml_task.get('name') != ruamel_task.get('name'):
-            raise RuntimeError('Error in matching skip comment to a task')
-        pyyaml_task['skipped_rules'] = _get_rule_skips_from_yaml(ruamel_task)
-
-    return pyyaml_data
-
-
-def _get_task_blocks_from_playbook(playbook):
-    """Return parts of playbook that contains tasks, and nested tasks.
-
-    :param playbook: playbook yaml from yaml parser.
-    :returns: list of task dictionaries.
-    """
-    PLAYBOOK_TASK_KEYWORDS = [
-        'tasks',
-        'pre_tasks',
-        'post_tasks',
-        'handlers',
-    ]
-
-    task_blocks = []
-    for play, key in product(playbook, PLAYBOOK_TASK_KEYWORDS):
-        task_blocks.extend(play.get(key, []))
-    return task_blocks
-
-
-def _get_tasks_from_blocks(task_blocks):
-    """Get list of tasks from list made of tasks and nested tasks."""
-    NESTED_TASK_KEYS = [
-        'block',
-        'always',
-        'rescue',
-    ]
-
-    def get_nested_tasks(task):
-        return (
-            subtask
-            for k in NESTED_TASK_KEYS if k in task
-            for subtask in task[k]
-        )
-
-    for task in task_blocks:
-        for sub_task in get_nested_tasks(task):
-            yield sub_task
-        yield task
-
-
-def _get_rule_skips_from_yaml(yaml_input):
-    """Travese yaml for comments with rule skips and return list of rules."""
-    def traverse_yaml(obj):
-        yaml_comment_obj_strs.append(str(obj.ca.items))
-        if isinstance(obj, dict):
-            for key, val in obj.items():
-                if isinstance(val, (dict, list)):
-                    traverse_yaml(val)
-        elif isinstance(obj, list):
-            for e in obj:
-                if isinstance(e, (dict, list)):
-                    traverse_yaml(e)
-        else:
-            return
-
-    yaml_comment_obj_strs = []
-    traverse_yaml(yaml_input)
-
-    rule_id_list = []
-    for comment_obj_str in yaml_comment_obj_strs:
-        for line in comment_obj_str.split('\\n'):
-            rule_id_list.extend(get_rule_skips_from_line(line))
-
-    return rule_id_list
-
-
-def get_rule_skips_from_line(line):
-    rule_id_list = []
-    if '# noqa' in line:
-        noqa_text = line.split('# noqa')[1]
-        rule_id_list = noqa_text.split()
-    return rule_id_list
-
-
-def normpath(path):
-    """
-    Normalize a path in order to provide a more consistent output.
-
-    Currently it generates a relative path but in the future we may want to
-    make this user configurable.
-    """
-    # convertion to string in order to allow receiving non string objects as
-    # arguments which would have failed under python2
-    return os.path.relpath(str(path))
-
-
-def is_playbook(filename):
+def is_playbook(filename: str) -> bool:
     """
     Check if the file is a playbook.
 
@@ -745,33 +644,69 @@ def is_playbook(filename):
     }
 
     # makes it work with Path objects by converting them to strings
-    if not isinstance(filename, six.string_types):
+    if not isinstance(filename, str):
         filename = str(filename)
 
     try:
         f = parse_yaml_from_file(filename)
     except Exception as e:
-        print(
-            "Warning: Failed to load %s with %s, assuming is not a playbook."
-            % (filename, e))
+        _logger.warning(
+            "Failed to load %s with %s, assuming is not a playbook.",
+            filename, e)
     else:
         if (
             isinstance(f, AnsibleSequence) and
+            hasattr(f, 'keys') and
             playbooks_keys.intersection(next(iter(f), {}).keys())
         ):
             return True
     return False
 
 
-def get_playbooks_and_roles(options=None):
+def get_yaml_files(options: Namespace) -> dict:
+    """Find all yaml files."""
+    # git is preferred as it also considers .gitignore
+    git_command = ['git', 'ls-files', '*.yaml', '*.yml']
+    _logger.info("Discovering files to lint: %s", ' '.join(git_command))
+
+    out = None
+
+    try:
+        out = subprocess.check_output(
+            git_command,
+            stderr=subprocess.STDOUT,
+            universal_newlines=True
+        ).split()
+    except subprocess.CalledProcessError as exc:
+        _logger.warning(
+            "Failed to discover yaml files to lint using git: %s",
+            exc.output.rstrip('\n')
+        )
+    except FileNotFoundError as exc:
+        if options.verbosity:
+            _logger.warning(
+                "Failed to locate command: %s", exc
+            )
+
+    if out is None:
+        out = [
+            os.path.join(root, name)
+            for root, dirs, files in os.walk('.')
+            for name in files
+            if name.endswith('.yaml') or name.endswith('.yml')
+        ]
+
+    return OrderedDict.fromkeys(sorted(out))
+
+
+# FIXME: drop noqa once this function is made simpler
+# Ref: https://github.com/ansible/ansible-lint/issues/744
+def get_playbooks_and_roles(options=None) -> List[str]:  # noqa: C901
     """Find roles and playbooks."""
     if options is None:
         options = {}
 
-    # git is preferred as it also considers .gitignore
-    files = OrderedDict.fromkeys(sorted(subprocess.check_output(
-        ["git", "ls-files", "*.yaml", "*.yml"],
-        universal_newlines=True).split()))
+    files = get_yaml_files(options)
 
     playbooks = []
     role_dirs = []
@@ -791,9 +726,16 @@ def get_playbooks_and_roles(options=None):
 
     for p in map(Path, files):
 
-        if any(str(p).startswith(file_path) for file_path in options.exclude_paths):
+        try:
+            for file_path in options.exclude_paths:
+                if str(p.resolve()).startswith(str(file_path)):
+                    raise FileNotFoundError(
+                        f'File {file_path} matched exclusion entry: {p}')
+        except FileNotFoundError as e:
+            _logger.debug('Ignored %s due to: %s', p, e)
             continue
-        elif (next((i for i in p.parts if i.endswith('playbooks')), None) or
+
+        if (next((i for i in p.parts if i.endswith('playbooks')), None) or
                 'playbook' in p.parts[-1]):
             playbooks.append(normpath(p))
             continue
@@ -816,15 +758,27 @@ def get_playbooks_and_roles(options=None):
         if p.parts[-1].startswith('.'):
             continue
 
-        if is_playbook(p):
+        if is_playbook(str(p)):
             playbooks.append(normpath(p))
             continue
 
-        if options.verbosity:
-            print('Unknown file type: %s' % normpath(p))
+        _logger.info('Unknown file type: %s', normpath(p))
 
-    if options.verbosity:
-        print('Found roles: ' + ' '.join(role_dirs))
-        print('Found playbooks: ' + ' '.join(playbooks))
+    _logger.info('Found roles: %s', ' '.join(role_dirs))
+    _logger.info('Found playbooks: %s', ' '.join(playbooks))
 
     return role_dirs + playbooks
+
+
+def expand_path_vars(path: str) -> str:
+    """Expand the environment or ~ variables in a path string."""
+    path = path.strip()
+    path = os.path.expanduser(path)
+    path = os.path.expandvars(path)
+    return path
+
+
+def expand_paths_vars(paths: List[str]) -> List[str]:
+    """Expand the environment or ~ variables in a list."""
+    paths = [expand_path_vars(p) for p in paths]
+    return paths
